using System;
using System.Numerics;
using SlatedGameToolkit.Framework.Utilities;

namespace SlatedGameToolkit.Framework.Graphics.Render
{
    public class Camera {
        private bool viewUpdated = true, projectionUpdated = true, orthographic;
        public virtual Vector3 Up { get; set; } = Vector3.UnitY;
        private Vector3 position, lookAt;
        private float nearField = 0.01f, farField = 100f;
        private float width, height;
        private Matrix4x4 lookAtMatrix = Matrix4x4.Identity;
        private Matrix4x4 view = Matrix4x4.Identity, projection;

        public event EventHandler projectionUpdateEvent;

        public bool Orthographic {
            get {
                return orthographic;
            } 
            
            set {
                orthographic = value;
                projectionUpdated = true;
            }
        }

        public Vector3 Position {
            get {
                return position;
            }
            set {
                this.position = value;
                viewUpdated = true;
            }
        }

        public Vector3 CameraFront {
            get {
                return (LookAt - Position).NormalizeSafe(default(Vector3));
            }
            set {
                LookAt = (position + value.NormalizeSafe(default(Vector3)));
            }
        }

        public Vector3 LookAt {
            get {
                return lookAt;
            }
            set {
                this.lookAt = value;
                viewUpdated = true;
            }
        }

        public virtual float NearField {
            get {
                return nearField;
            }

            set {
                nearField = value;
                projectionUpdated = true;
            }
        }

        public virtual float FarField {
            get {
                return farField;
            }

            set {
                farField = value;
                projectionUpdated = true;
            }
        }

        public virtual float Width {
            get {
                return width;
            }
            set {
                width = value;
                projectionUpdated = true;
            }
        }
        public virtual float Height {
            get {
                return height;
            }
            set {
                height = value;
                projectionUpdated = true;
            }
        }

        public Matrix4x4 ViewMatrix {
            get {
                if (viewUpdated) {
                    view = Matrix4x4.CreateLookAt(Position, LookAt, Up);
                    viewUpdated = false;
                }
                return view;
            }
        }

        public Matrix4x4 ProjectionMatrix {
            get {
                if (projectionUpdated) {
                    if (Orthographic) {
                        projection = Matrix4x4.CreateOrthographic(width, height, nearField, farField);
                    } else {
                        projection = Matrix4x4.CreatePerspective(width, height, nearField, farField);
                    }
                    projectionUpdated = false;
                }
                return projection;
            }
        }

        public Camera(float width, float height) {
            if (width <= 0) throw new ArgumentException("Width can't be less than or equal to 0.");
            if (height <= 0) throw new ArgumentException("Height can't be less than or equal to 0.");
            this.Width = width;
            this.Height = height;
            this.Position = new Vector3(0, 0, 3);
            this.CameraFront = -Vector3.UnitZ;
        }

        private void OnProjectionUpdated() {
            projectionUpdateEvent?.Invoke(this, new EventArgs());
        }
    }
}