using System;
using System.IO;
using System.Runtime.InteropServices;
using SDL2;
using SlatedGameToolkit.Framework.Exceptions;
using SlatedGameToolkit.Framework.Graphics.OpenGL;
using SlatedGameToolkit.Framework.Graphics.Textures;
using SlatedGameToolkit.Framework.Loader;

namespace SlatedGameToolkit.Framework.Loader
{
    public static class CommonLoaders
    {
        /// <summary>
        /// Loads a texture using SDL2's image library.
        /// Therefore, technically, this function should be able to laod any format SDL2 Image can load.
        /// </summary>
        /// <param name="path">The path of the texture to load.</param>
        /// <param name="glContext">The OpenGL context the texture is to be associated with. May be null.</param>
        /// <returns>An OpenGL Texture associated with the given context.</returns>
        public static Texture LoadTexture(string path, GLContext glContext)
        {
            IntPtr ptr = SDL_image.IMG_Load(path);
            if (ptr.Equals(IntPtr.Zero)) throw new FrameworkSDLException("Could not find texture file at path: " + path);
            SDL.SDL_Surface surface = Marshal.PtrToStructure<SDL.SDL_Surface>(ptr);
            SDL.SDL_PixelFormat pixelFormat = Marshal.PtrToStructure<SDL.SDL_PixelFormat>(surface.format);
            if (pixelFormat.format != SDL.SDL_PIXELFORMAT_ABGR8888) {
                GameEngine.Logger.Warning(string.Format("Texture \"{0}\" is of format {1}. Converting to ABGR8888. Consider saving files in such a data format for faster load times.", path, SDL.SDL_GetPixelFormatName(pixelFormat.format).Remove(0, 16)));
                IntPtr convertedPtr = SDL.SDL_ConvertSurfaceFormat(ptr, SDL.SDL_PIXELFORMAT_ABGR8888, 0);
                if (convertedPtr == null) throw new OptionalSDLException();
                SDL.SDL_FreeSurface(ptr);
                ptr = convertedPtr;
                surface = Marshal.PtrToStructure<SDL.SDL_Surface>(ptr);
            }

            byte[] data = new byte[surface.pitch * surface.h];
            Marshal.Copy(surface.pixels, data, 0, data.Length);
            
            TextureData textureData = new TextureData(surface.w, surface.h, data);

            Texture useable = new Texture(textureData, glContext);
            SDL.SDL_FreeSurface(ptr);
            return useable;
        }
    }
}